#!/usr/bin/env ruby

#
# == http_traceroute: Show redirects from start URL along with
# other useful information
# 
# == Usage
#
# http_traceroute Robin Wood (robin@digi.ninja) (http://digi.ninja)
# 
# Usage: http_traceroute [OPTION] ... URL
# 	--help, -h, -?: show help
# 	--max-redirects, -m X: maximum number of redirects (default 10)
# 	--large-body, -b X: alert if the body on a redirect is larger than X (default 500b)
#	--cookie-jar, -c X: use the specified cookie jar"
#	--ua, -u user-agent X: useragent to send, default "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0"
#
#
#	Proxy Support
#	--proxy-host: proxy host
#	--proxy-port: proxy port, default 8080
#	--proxy-username: username for proxy, if required
#	--proxy-password: password for proxy, if required
# 
# 	URL: The URL to start on
#
# Author:: Robin Wood (robin@digi.ninja)
# Copyright:: Copyright (c) Robin Wood 2014
# Licence:: Creative Commons Attribution-Share Alike 2.0
#

require "net/http"
require "net/https"
require "uri"
require "getoptlong"

@cookie_jar_enabled = true

@proxy_host = nil
@proxy_port = nil
@proxy_username = nil
@proxy_password = nil

begin
	require "json"
	require "./cookiejar/lib/cookiejar.rb"
	require "./mycookiejar"
rescue LoadError => e
	puts e.inspect
	@cookie_jar_enabled = false
end

VERSION = "1.1"
if @cookie_jar_enabled
	@cookie_jar = CookieJar::Jar.new()
end

def puts_msg_with_header (msg)
	puts"http_traceroute #{VERSION} Robin Wood (robin@digi.ninja) (http://digi.ninja)\n"
	puts
	puts msg
end

class String
	def is_numeric?
		Integer self rescue false
	end

	# Wrap the string to specified length
	def wrap(width = 80)
		scan(/\S.{0,#{width-2}}\S(?=\s|$)|\S+/)
	end
end

def usage
	# broken
	#--timeout, -t X: timeout trying to connect (default 60 seconds, miniumum seems to be 15 seconds)
	puts "Usage: http_traceroute [OPTION] ... URL
	--help, -h, -?: show help
	--max-redirects, -m X: maximum number of redirects (default 10)
	--large-body, -b X: alert if the body on a redirect is larger than X (default 500b)
	--ua, -u user-agent: useragent to send, default:
		Mozilla/5.0 (Windows NT 6.3; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0
"
	if @cookie_jar_enabled then
		puts "	--cookie-jar, -c X: use the specified cookie jar"
	end
puts "
	Proxy Support
	--proxy-host: proxy host
	--proxy-port: proxy port, default 8080
	--proxy-username: username for proxy, if required
	--proxy-password: password for proxy, if required
	
	URL: The URL to start on
"
	if not @cookie_jar_enabled then
		puts
		puts "CookieJar gem not installed so cookie jar feature not enabled. To enable it use:"
		puts "\tgem install cookiejar"
	end
end

def print_cookies (uri, headers)
	print_lined "Cookies"

	tmp_jar = CookieJar::Jar.new()
	tmp_jar.set_cookies_from_headers uri, headers.to_hash

	cookies = tmp_jar.to_a
	cookies.each do |cookie|
		puts cookie.for_display
	end

	puts
end

def print_headers (headers)
	print_lined "Headers"

	headers.to_hash.each do |key, values|
		values.each do |value|
			puts key + ": " + value
		end
	end
	puts
end

def print_lined (text)
	puts "="*text.length
	puts text
	puts "="*text.length
end

def get_page (url, large_body_size, limit = 10)
	if limit == 0
		print_lined "Redirection limit hit, maybe a loop, aborting"
		return
	end

	puts "*"*60
	puts "Requesting: " + url

	uri = URI.parse(url)
	ssl = false

	http = Net::HTTP.new(uri.host, uri.port, @proxy_host, @proxy_port, @proxy_username, @proxy_password)
	case uri.scheme
		when "https"
			ssl = true
			http.use_ssl = true
		when "http"
			http.use_ssl = false
		else
			raise "Unknown scheme (protocol): " + uri.scheme
	end

	request = Net::HTTP::Get.new(uri.request_uri, {"User-Agent" => @ua})

	unless @cookie_jar.nil?
		cookies = @cookie_jar.get_cookies uri
		if cookies.count > 0
			cookie_string = cookies.to_a.join(" ; ")
			request['cookie'] = cookie_string
			puts "Sending cookies: " + cookie_string
		end
	end

	x509 = nil
	response = nil
	begin
		http.start do
	# Seem to be ignored
	#http.read_timeout = @timeout
	#http.open_timeout = @timeout
	#http.ssl_timeout = @timeout
			response = http.request(request)

			# tried checking like this but it fails
			# if http.use_ssl
			if ssl
				x509 = http.peer_cert()
			end
		end
	rescue Errno::ECONNREFUSED
		puts
		if @proxy_host.nil?
			print_lined "Error: Connection refused"
		else
			print_lined "Error: Connection refused, check the proxy settings"
		end
		exit
	rescue Net::OpenTimeout, Net::ReadTimeout, Errno::ETIMEDOUT
		puts
		print_lined "Error: Timeout connecting to the site"
		exit
	rescue Errno::EINVAL
		puts
		print_lined "Error: There was a problem connecting to the site"
		exit
	rescue SocketError => e
		puts
		print_lined "Error: There was a problem connecting to the site"
		exit
	rescue OpenSSL::SSL::SSLError => e
		puts "*** Invalid SSL certificate, trying again ignoring certificate ***"
		puts

		http.verify_mode = OpenSSL::SSL::VERIFY_NONE
		begin
			http.start do
				response = http.request(request)
				if ssl
					x509 = http.peer_cert()
				end
			end
			rescue OpenSSL::SSL::SSLError => e
				print_lined "Error: Still having certificate errors, aborting"
				exit
			rescue Errno::ECONNREFUSED
				puts
				if @proxy_host.nil?
					print_lined "Error: Connection refused"
				else
					print_lined "Error: Connection refused, check the proxy settings"
				end
				exit
			rescue Net::OpenTimeout, Net::ReadTimeout, Errno::ETIMEDOUT
				puts
				print_lined "Error: Timeout connecting to the site"
				exit
			rescue Errno::EINVAL
				puts
				print_lined "Error: There was a problem connecting to the site"
				exit
			rescue SocketError => e
				puts
				print_lined "Error: There was a problem connecting to the site"
				exit
				end
			rescue => e
				puts "There was a problem connecting to the site"
				puts e
				exit
	end

	unless @cookie_jar.nil?
		@cookie_jar.set_cookies_from_headers uri, response.to_hash
	end

	puts "Response code: " + response.code
	puts

	print_headers(response)
	unless @cookie_jar.nil?
		print_cookies(uri, response)
	end

	if not x509.nil?
		print_lined "Certificate Information"
		puts "Subject: " + x509.subject.to_s
		puts "Version: " + x509.version.to_s
		puts "Issuer: " + x509.issuer.to_s
		puts "Not valid before: " + x509.not_before.to_s
		puts "Not valid after: " + x509.not_after.to_s
		puts
	end

	case response
		# This could be just an if redirect else return
		# but this way there is the option to do different
		# things based on the return code
		when Net::HTTPSuccess, Net::HTTPNotFound
			print_lined "End of the line"
			puts
		when Net::HTTPRedirection then
			print_lined "Redirecting"
			if response.header.key? "content-length"
				length_str = response.header["content-length"]
				if length_str.is_numeric?
					if length_str.to_i > large_body_size
						puts "*** Large body on redirect - length: " + length_str + " ***"
					end
				end
			end
			if response.header.key? "location"
				location = response['location']
				puts "Location: " + location
				uri_redirect = URI.parse(location)
				case uri_redirect
					when URI::HTTP
						new_url = location
					when URI::Generic
						if location[0] == "/"
							new_url = uri.scheme + "://" + uri.host + location
						else
							if uri.path[-1] == "/"
								dir = uri.path
							else
								dir = File.dirname uri.path
								if dir[-1] != "/"
									dir += "/"
								end
							end
							new_url = uri.scheme + "://" + uri.host + dir + location
						end
					else
						raise "Odd redirect"
				end

				puts "Redirecting to: " + new_url
				puts

				return get_page(new_url, large_body_size, limit -1)
			else
				raise "No location field"
			end
		else
			puts "End of the line"
	end
end

large_body_size = 500
max_redirects = 10
url = nil
cookie_jar_file = nil
@ua = "Mozilla/5.0 (Windows NT 6.3; WOW64; rv:33.0) Gecko/20100101 Firefox/33.0"
@timeout = 60

opts = GetoptLong.new(
	[ '--help', '-h', "-?", GetoptLong::NO_ARGUMENT ],
	[ '--max-redirects', "-m" , GetoptLong::REQUIRED_ARGUMENT ],
#	[ '--timeout-redirects', "-t" , GetoptLong::REQUIRED_ARGUMENT ],
	[ '--large-body', "-b" , GetoptLong::REQUIRED_ARGUMENT ],
	[ '--cookie-jar', "-c" , GetoptLong::REQUIRED_ARGUMENT ],
	[ '--ua', "-u" , GetoptLong::REQUIRED_ARGUMENT ],
	[ '--proxy-host', GetoptLong::REQUIRED_ARGUMENT ],
	[ '--proxy-port', GetoptLong::REQUIRED_ARGUMENT ],
	[ '--proxy-username', GetoptLong::REQUIRED_ARGUMENT ],
	[ '--proxy-password', GetoptLong::REQUIRED_ARGUMENT ],
)

puts "http_traceroute #{VERSION} Robin Wood (robin@digi.ninja) (http://digi.ninja)"
puts

begin
	opts.each do |opt, arg|
		case opt
			when '--help'
				usage
				exit
			when "--cookie-jar"
				cookie_jar_file = arg
			when "--large-body"
				if arg.is_numeric?
					large_body_size = arg.to_i
					if large_body_size <= 0
						puts_msg_with_header("Please enter a positive size for the large body value")
						exit 1
					end
				else
					puts_msg_with_header("Large body size wasn't numeric")
					exit 1
				end
			when "--timeout"
				if arg.is_numeric?
					@timeout = arg.to_i
					if @timeout <= 0
						puts_msg_with_header("Please enter a positive number for the timeout")
						exit 1
					end
				else
					puts_msg_with_header("Invalid timeout")
					exit 1
				end
			when "--max-redirects"
				if arg.is_numeric?
					max_redirects = arg.to_i
					if max_redirects <= 0
						puts_msg_with_header("Please enter a positive number of redirects")
						exit 1
					end
				else
					puts_msg_with_header("Invalid number of redirects")
					exit 1
				end
			when '--ua'
				@ua=arg
			when "--proxy-password"
				@proxy_password = arg.to_s
			when "--proxy-username"
				@proxy_username = arg.to_s
			when "--proxy-host"
				@proxy_host = arg.to_s
			when "--proxy-port"
				unless arg.is_numeric?
					print_lined "Error: Proxy port must be numeric"
					exit
				end
				@proxy_port = arg.to_i
				if @proxy_port < 0
					print_lined "Error: Proxy port must be positive"
					exit
				end
		end
	end
rescue GetoptLong::MissingArgument
	puts
	usage
	exit
rescue GetoptLong::InvalidOption => e
	puts
	usage
	exit
rescue => e
	puts "Something went wrong, please report it to robin@digi.ninja along with these messages:"
	puts
	puts e.message
	puts
	puts e.class.to_s
	puts
	puts "Backtrace:"
	puts e.backtrace
	puts
	usage
	exit 1
end

# Default the proxy port if host set but no port given
if !@proxy_host.nil? and @proxy_port.nil?
	@proxy_port = 8080
end

url = ARGV.shift

unless @proxy_host.nil?
	puts "Using proxy at " + @proxy_host + ":" + @proxy_port.to_s
end

if url.nil?
	puts "You didn't provide a URL to test"
	exit
end

if url !~ /^https?:\/\//i
	print_lined "Error: Please specify a protocol, http or https"
	puts
	exit
end

unless cookie_jar_file.nil?
	if File.exists? cookie_jar_file
		begin
			@cookie_jar.load cookie_jar_file
		rescue JSON::ParserError
			print_lined "Error: Cookie jar could not be processed, aborting"
			puts
			exit
		end
	end
end

puts "*"*60
puts "Starting at: " + url
response = get_page(url, large_body_size, max_redirects)

unless cookie_jar_file.nil?
	begin
		@cookie_jar.save(cookie_jar_file)
		print_lined "Cookies saved to " + cookie_jar_file
		puts
	rescue
		print_lined "Error: Cookie jar could not be saved"
		puts
		exit
	end
end
